module Towel
  # Loads and merges configuration settings for Towel.
  class Configuration
    # Acceptable filenames in which to define Towel settings. These will be
    # read in the defined order.
    FILENAMES = %w[Towel.toml towel.toml .Towel.toml .towel.toml].freeze

    # Read the configuration from files and environment variables.
    def self.read
      # Load configuration from files, starting in the current working directory
      # and working upwards in the filesystem hierarchy until no configuration
      # files are found.
      files = []

      directory = Dir.pwd
      while true
        FILENAMES.each do |filename|
          possible_file = File.join(directory, filename)
          files << possible_file if File.exist?(possible_file)
        end

        parent = File.expand_path("..", directory)
        if directory == parent
          break
        else
          directory = parent
        end
      end

      # Combine all discovered files into one configuration.
      config = files
        .reverse
        .map { |file| TOML.load_file(file) }
        .inject({}) do |acc, toml|
          acc.merge(toml) do |key, old, new|
            case old
            when Hash then old.merge(new)
            else
              new
            end
          end
        end

      # Override with environment variables where appropriate.
      override = ->(env, section, key) do
        if ENV.key?(env)
          section_values = config[section] || {}
          section_values[key] = ENV[env]
          config[section] = section_values
        end
      end
      override.call("TOWEL_API_KEY", "auth", "api_key")
      override.call("TOWEL_ACCOUNT", "auth", "account")
      override.call("TOWEL_ADDRESS", "collector", "address")

      new(config)
    end

    # Creates a new configuration from a Hash.
    def initialize(hash = {})
      @config = {
        "auth" => {
          "account" => nil,
          "api_key" => nil
        },
        "collector" => {
          "address" => "towel.dev",
          "organization" => nil,
          "project" => nil
        },
        "labels" => {}
      }
      @config.merge!(hash)
      @config.freeze
    end

    # Retrieves a configuration section ("auth", "collector", "labels", etc).
    def [](key)
      @config[key]
    end
  end
end
