module Towel
  class Log
    # Logs must be associated with an invocation.
    NAME_FORMAT = /invocations\/[^\/]+\/logs\/[^\/]+/

    # How long to wait for more log entries before calling `AppendLog`. This
    # tries to batch log lines together as opposed to making a call for each
    # log line individually.
    FLUSH_INTERVAL_SECONDS = 0.1

    # How many log lines to send at a time. If there are more than this, they
    # will be broken up into several `AppendLog` calls.
    FLUSH_COUNT = 100

    attr_reader :name

    def initialize(name, stub)
      unless name =~ NAME_FORMAT
        raise ArgumentError, "Log name #{name} is invalid"
      end
      raise ArgumentError, "Stub must be provided" unless stub
      @name = name
      @stub = stub
      @entries = Queue.new
      @entry_id = Concurrent::AtomicFixnum.new(-1)
      @flush_mutex = Mutex.new
      @flush_timer_mutex = Mutex.new
      @flush_timer = nil

      # Create the log
      log = Towel::V1alpha::Log.new
      log.name = @name
      request = Towel::V1alpha::CreateLogRequest.new
      request.log = log
      @stub.create_log(request)
    end

    def context
      Thread.current["towel_context_#{object_id}"]
    end

    def context=(context)
      Thread.current["towel_context_#{object_id}"] = context
    end

    def <<(line)
      entry = Towel::V1alpha::LogEntry.new
      entry.entry_id = @entry_id.increment
      entry.log_time = Time.now.utc
      entry.parent = context if context
      entry.contents = line

      @entries << entry

      # Set up a flush to happen momentarily, allowing other lines to possibly
      # be flushed together with this one.
      @flush_timer_mutex.synchronize do
        unless @flush_timer
          @flush_timer = Concurrent::ScheduledTask.execute(
            FLUSH_INTERVAL_SECONDS
          ) { flush }
        end
      end
    end

    def close
      flush
      Thread.current["towel_context_#{object_id}"] = nil
    end

    # Ensure that entries get written out to Towel.
    def flush
      @flush_mutex.synchronize do
        # Acknowledge the flush timer by unsetting it, if it exists.
        @flush_timer_mutex.synchronize do
          @flush_timer.cancel if @flush_timer
          @flush_timer = nil
        end

        until @entries.empty?
          request = Towel::V1alpha::AppendLogRequest.new
          request.name = @name
          i = 0
          while i < FLUSH_COUNT
            i += 1
            begin
              entry = @entries.pop(true)
            rescue ThreadError # Raised if the queue is empty
              break
            end
            request.entries << entry
          end
          @stub.append_log(request)
        end
      end
    end
  end
end
