module Towel
  # An IO-like object that logs to Towel. Sits on top of Log. This relationship
  # is inverted compared to most IO class structures, but log lines/entries are
  # the fundamental unit of Towel. This class is merely an adapter to collect
  # logs from IO streams like `$stdout` and `$stderr`.
  class LogIO
    def initialize(log)
      raise ArgumentError, "Log must be specified" unless log
      @log = log
      @buffer = ""
      @write_mutex = Mutex.new
    end

    def context=(context)
      @log.context = context
    end

    def putc(c)
      if c.is_a?(String)
        write(c[0])
      else
        write(c.chr)
      end
      c
    end

    def puts(*args)
      if args.empty?
        write($/)
      else
        args.each do |arg|
          if arg.end_with?($/)
            write(arg)
          else
            write(arg, $/)
          end
        end
      end
      nil
    end

    def print(*args)
      if args.empty?
        write($_)
      else
        if $\.nil?
          write(args.join($,))
        else
          write(args.join($,), $\)
        end
      end
      nil
    end

    def printf(*args)
      write(sprintf(*args))
      nil
    end

    def <<(obj)
      write(obj)
      self
    end

    def write(*args)
      @write_mutex.synchronize do
        bytes = 0
        args.each do |arg|
          arg = arg.to_s
          @buffer << arg
          bytes += arg.bytesize
        end

        remaining = ""
        @buffer.each_line do |line|
          if line.end_with?($/)
            @log << line
          else
            remaining = line
          end
        end
        @buffer = remaining

        bytes
      end
    end

    def flush
      @log.flush
    end

    def close
      @log.close
    end
  end
end
